#pragma once

#include <polymesh/cursors.hh>
#include <typed-geometry/tg.hh>
#include <vector>

namespace polymesh
{
/// "Hierarchical additive poisson disk sampling" [Dieckmann and Klein]
/// Returns a list of Poisson disk sampling of the mesh's vertices with increasing resolution.
/// The union of samplings is also a Poisson sampling (i.e. no points closer to one another than radius of finest sampling).
std::pair < std::vector<std::vector<polymesh::vertex_handle>>, std::vector<tg::aabb3>>
create_hierarchical_poisson_sampling(polymesh::vertex_attribute<tg::pos3> const& position,
                                                                                       float finest_sampling_radius,
                                                                                       int num_layers);

// user does not have to specify the number of lods - we can try and take care of it, assuming leaf *splat radius* was given!
std::pair<std::vector<std::vector<polymesh::vertex_handle>>, std::vector<tg::aabb3>> create_hierarchical_poisson_sampling(polymesh::vertex_attribute<tg::pos3> const& position,
                                                                                                                          float leaf_splat_radius);
}
